#include "stdafx.h"
#include "CBData.h"
#include "GlobalPtr.h"
#include "util.h"
#include "ClipMM.h"
#include "CopyingDialog.h"
#include "NFile.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

using namespace Gdiplus;


#define PFT_ITEM(cf) {	cf,	L#cf	}
static map<UINT, CString> mapPredefinedFormatTable =
{
	PFT_ITEM(CF_BITMAP),
	PFT_ITEM(CF_DIB),
	PFT_ITEM(CF_DIBV5),
	PFT_ITEM(CF_DIF),

	PFT_ITEM(CF_DSPBITMAP),
	PFT_ITEM(CF_DSPENHMETAFILE),
	PFT_ITEM(CF_DSPMETAFILEPICT),
	PFT_ITEM(CF_DSPTEXT),

	PFT_ITEM(CF_ENHMETAFILE),
	PFT_ITEM(CF_HDROP),
	PFT_ITEM(CF_LOCALE),
	PFT_ITEM(CF_METAFILEPICT),
	PFT_ITEM(CF_OEMTEXT),

	PFT_ITEM(CF_OWNERDISPLAY),
	PFT_ITEM(CF_PALETTE),
	PFT_ITEM(CF_PENDATA),
	PFT_ITEM(CF_RIFF),
	PFT_ITEM(CF_SYLK),

	PFT_ITEM(CF_TEXT),
	PFT_ITEM(CF_TIFF),
	PFT_ITEM(CF_UNICODETEXT),
	PFT_ITEM(CF_WAVE),

};

#define RFT_ITEM(cf) {	cf,	L#cf	}
static map<CString, CString> mapRegisteredFormatTable =
{
	RFT_ITEM(CFSTR_SHELLIDLIST),
	RFT_ITEM(CFSTR_SHELLIDLISTOFFSET),
	RFT_ITEM(CFSTR_NETRESOURCES),
	RFT_ITEM(CFSTR_FILEDESCRIPTORA),
	RFT_ITEM(CFSTR_FILEDESCRIPTORW),
	RFT_ITEM(CFSTR_FILECONTENTS),
	RFT_ITEM(CFSTR_FILENAMEA),
	RFT_ITEM(CFSTR_FILENAMEW),
	RFT_ITEM(CFSTR_PRINTERGROUP),
	RFT_ITEM(CFSTR_FILENAMEMAPA),
	RFT_ITEM(CFSTR_FILENAMEMAPW),
	RFT_ITEM(CFSTR_SHELLURL),
	RFT_ITEM(CFSTR_INETURLA),
	RFT_ITEM(CFSTR_INETURLW),
	RFT_ITEM(CFSTR_PREFERREDDROPEFFECT),
	RFT_ITEM(CFSTR_PERFORMEDDROPEFFECT),
	RFT_ITEM(CFSTR_PASTESUCCEEDED),
	RFT_ITEM(CFSTR_INDRAGLOOP),
	RFT_ITEM(CFSTR_MOUNTEDVOLUME),
	RFT_ITEM(CFSTR_PERSISTEDDATAOBJECT),
	RFT_ITEM(CFSTR_TARGETCLSID),
	RFT_ITEM(CFSTR_LOGICALPERFORMEDDROPEFFECT),
	RFT_ITEM(CFSTR_AUTOPLAY_SHELLIDLISTS),
	RFT_ITEM(CFSTR_UNTRUSTEDDRAGDROP),
	RFT_ITEM(CFSTR_FILE_ATTRIBUTES_ARRAY),
	RFT_ITEM(CFSTR_INVOKECOMMAND_DROPPARAM),
	RFT_ITEM(CFSTR_SHELLDROPHANDLER),
	RFT_ITEM(CFSTR_DROPDESCRIPTION),
	RFT_ITEM(CFSTR_MIME_TEXT),
};

BOOL GetClipboardFormatNameString(UINT nCF, CString& strName, BOOL& bRegistered)
{
	// `ς݂H
	auto iter = mapPredefinedFormatTable.find(nCF);
	if (iter != mapPredefinedFormatTable.end()) 
	{
		bRegistered = FALSE;
		strName = iter->second;
		return TRUE;
	}

	// `ς݂͈̔͂H
	if (nCF >= CF_GDIOBJFIRST && nCF <= CF_GDIOBJLAST)
	{
		bRegistered = FALSE;
		strName.Format(L"CF_GDIOBJFIRST+%u", nCF - CF_GDIOBJFIRST);
		return TRUE;
	}
	if (nCF >= CF_PRIVATEFIRST && nCF <= CF_PRIVATELAST)
	{
		bRegistered = FALSE;
		strName.Format(L"CF_PRIVATEFIRST+%u", nCF - CF_PRIVATEFIRST);
		return TRUE;
	}

	// registered formatH
	static const int cchFormatNameBuf = 65535;
	vector<WCHAR> szFormatName(cchFormatNameBuf);
	if (GetClipboardFormatName(nCF, szFormatName.data(), cchFormatNameBuf))
	{
		auto iter = mapRegisteredFormatTable.find(szFormatName.data());
		if (iter != mapRegisteredFormatTable.end())
		{
			bRegistered = TRUE;
			strName = iter->second;
			return TRUE;
		}
		else
		{
			bRegistered = TRUE;
			strName = CString(L"\"") + szFormatName.data() + L"\"";
			return TRUE;
		}
	}

	// sB
	bRegistered = FALSE;
	strName.Format(L"%u", nCF);
	return TRUE;
}

CString GetClipboardFormatNameString(UINT nCF)
{
	CString strName;
	BOOL bRegistered = FALSE;
	GetClipboardFormatNameString(nCF, strName, bRegistered);
	return strName;
}

CString FORMATETCAspect2String(DWORD dwAspect)
{
	CStringArray arrayAspect;
	if (dwAspect & DVASPECT_CONTENT)
		arrayAspect.Add(L"CONTENT");
	if (dwAspect & DVASPECT_THUMBNAIL)
		arrayAspect.Add(L"THUMBNAIL");
	if (dwAspect & DVASPECT_ICON)
		arrayAspect.Add(L"ICON");
	if (dwAspect & DVASPECT_DOCPRINT)
		arrayAspect.Add(L"DOCPRINT");
	if (dwAspect & ~(DVASPECT_CONTENT | DVASPECT_THUMBNAIL | DVASPECT_ICON | DVASPECT_DOCPRINT))
	{
		CString str;
		str.Format(L"0x%x?", dwAspect);
		arrayAspect.Add(str);
	}
	return StringFromStringArray(arrayAspect, L" ");
}

CString FORMATETCTymed2String(DWORD tymed)
{
	CStringArray arrayTymed;
	if (tymed & TYMED_HGLOBAL)
		arrayTymed.Add(L"HGLOBAL");
	if (tymed & TYMED_FILE)
		arrayTymed.Add(L"FILE");
	if (tymed & TYMED_ISTREAM)
		arrayTymed.Add(L"ISTREAM");
	if (tymed & TYMED_ISTORAGE)
		arrayTymed.Add(L"ISTORAGE");
	if (tymed & TYMED_GDI)
		arrayTymed.Add(L"GDI");
	if (tymed & TYMED_MFPICT)
		arrayTymed.Add(L"MFPICT");
	if (tymed & TYMED_ENHMF)
		arrayTymed.Add(L"ENHMF");
	return StringFromStringArray(arrayTymed, L" ");
}

CString FORMATETC2String(const FORMATETC& fmtetc)
{
	CStringArray arrayString;
	// tH[}bg
	arrayString.Add(GetClipboardFormatNameString(fmtetc.cfFormat) + L" (" + UnsignedDecimalString(fmtetc.cfFormat) + L")");

	// Aspect
	arrayString.Add(CString(L"Aspect=") + FORMATETCAspect2String(fmtetc.dwAspect));

	// index
	{
		CString str;
		str.Format(L"Index=%d", fmtetc.lindex);
		arrayString.Add(str);
	}

	// tymed
	{
		arrayString.Add(CString(L"Tymed=") + FORMATETCTymed2String(fmtetc.tymed));
	}
	return StringFromStringArray(arrayString, L" ");	
}

CString FORMATETC2ShortString(const FORMATETC& fmtetc)
{
	CString strResult = GetClipboardFormatNameString(fmtetc.cfFormat);

	if (fmtetc.dwAspect && fmtetc.dwAspect != DVASPECT_CONTENT)
	{
		strResult += L" ";
		strResult += FORMATETCAspect2String(fmtetc.dwAspect);
	}

	return strResult;
}

void FORMATETC2ShortStringArray(vector<FORMATETC>& arrayFormats, CStringArray& arrayString)
{
	arrayString.RemoveAll();

	for (int i = 0; i < (int)arrayFormats.size(); i++)
	{
		arrayString.Add(FORMATETC2ShortString(arrayFormats[i]));
	}
}

/////////////////////////////////////

CCBDataCreateContext::CCBDataCreateContext(COleDataObject* pDataObject, CCopyingDialog* pDlg)
{
	ASSERT(pDataObject);

	m_pDataObject = pDataObject;
	m_hPalette = NULL;
	m_bPaletteCopied = FALSE;
	m_cbTotal = 0;

	m_pDlg = pDlg;

	// pbgΎĂ
	FORMATETC fmtetcPalette = {	CF_PALETTE, NULL, DVASPECT_CONTENT, -1, TYMED_GDI };
	STGMEDIUM stgmedPalette = {0};
	if (pDataObject->GetData(CF_PALETTE, &stgmedPalette, &fmtetcPalette))
	{
		HPALETTE hPalette = (HPALETTE)stgmedPalette.hBitmap;
		if (hPalette)
		{
			// releasedeletêŃRs[Ďg
			m_hPalette = (HPALETTE)OleDuplicateData(hPalette, CF_PALETTE, 0);
			m_bPaletteCopied = TRUE;
		}
		ReleaseStgMedium(&stgmedPalette);
	}

}

CCBDataCreateContext::~CCBDataCreateContext()
{
	if (m_hPalette && m_bPaletteCopied)
	{
		DeleteObject(m_hPalette);
	}
}

void CCBDataCreateContext::SetName(const CString& strName)
{
	m_pDlg->SetName(strName);
}

void CCBDataCreateContext::SetName(FORMATETC& fmtetc)
{
	m_pDlg->SetName(FORMATETC2ShortString(fmtetc));
}


BOOL CCBDataCreateContext::IsCancelled()
{
	MSG message;
	while (PeekMessage(&message, NULL, NULL, NULL, TRUE))
	{
		if (m_pDlg->m_hWnd && m_pDlg->IsDialogMessage(&message))
		{
		}
		else
		{
			TranslateMessage(&message);
			DispatchMessage(&message);
		}
	}

	return m_pDlg->IsCancelled();
}

CString CCBDataCreateContext::GetAbortText()
{
	CString str = L"G[\t";

	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();
	ULONGLONG cbLimit = (ULONGLONG)pApp->m_nSubDirSizeLimit * 1024 * 1024;
	if (m_cbTotal > cbLimit)
	{
		str += L"TCYE𒴂";
	}
	str += L"擾𒆎~܂B";
	return str;
}

/////////////////////////////////////

IMPLEMENT_DYNAMIC(CCBData, CObject)

CCBData::CCBData()
	: m_cbMem(0)
	, m_pArrayChildren(NULL)
	, m_suitabletype(suitabletypeHex)
	, m_nCodePage(CP_CBDATA_UNKNOWN)
	, m_tymed(0)
	, m_bPlaceHolder(FALSE)
	, m_nIconType(icontypeFile)

{
	FORMATETC fmtetcInit = {0};
	fmtetcInit.lindex = -1;
	m_fmtetc = fmtetcInit;
}

CCBData::~CCBData()
{
	if (!m_strMemDataFileName.IsEmpty())
		DeleteFile(m_strMemDataFileName);
}

void CCBData::DeriveFrom(shared_ptr<CCBData> pBase)
{
	ASSERT(this->IsKindOf(pBase->GetRuntimeClass()));
	ASSERT(CString(pBase->GetRuntimeClass()->m_lpszClassName) == L"CCBData");

	// {NX̃IuWFNgSzŝĂ
	m_fmtetc = pBase->m_fmtetc;
	m_tymed = pBase->m_tymed;
	m_strType = pBase->m_strType;
	m_strFileName = pBase->m_strFileName;
	m_arrayProperties.Copy(pBase->m_arrayProperties);
	m_cbMem = pBase->m_cbMem;
	m_strMemDataFileName = pBase->m_strMemDataFileName;
	pBase->m_strMemDataFileName.Empty();// 폜Ȃ悤

	if (pBase->m_pArrayChildren)
	{
		m_pArrayChildren = pBase->m_pArrayChildren;
		pBase->m_pArrayChildren = NULL;
	}
	m_bPlaceHolder = pBase->m_bPlaceHolder;
	m_pShellFolder = pBase->m_pShellFolder;
	pBase->m_pShellFolder = NULL;
	m_pStorage = pBase->m_pStorage;
	pBase->m_pStorage = NULL;
}

bool CCBData::LoadMemData(vector<BYTE>& buf)
{
	if (!m_cbMem)
		return false;

	if (m_cbMem > SIZE_T_MAX)
	{
		return false;
	}

	if (m_strMemDataFileName.IsEmpty())
		return false;

	NFile file;
	if (file.OpenExistingFileToRead(m_strMemDataFileName))
	{
		BYTE* pMem = NULL;
		try
		{
			buf.resize((size_t)m_cbMem);
			pMem = buf.data();
		}
		catch (std::bad_alloc)
		{
			return false;
		}
		if (!pMem) 
		{
			return false;
		}
		size_t cbRemain = (size_t)m_cbMem;
		size_t nOffset = 0;
		while (cbRemain)
		{
			UINT cbToRead = (UINT)min(1024 * 1024 * 1024, cbRemain);
			UINT cbRead = file.Read(pMem + nOffset, cbToRead);
			if (cbToRead != cbRead)
			{
				break;
			}
			cbRemain -= cbToRead;
			nOffset += cbToRead;
		}
		if (cbRemain)
		{
			AfxMessageBox(L"ꎞt@Cǂ߂܂");
		}
		return true;
	}
	return false;
}

shared_ptr<CCBData> CCBData::Create(CCBDataCreateContext* pctx, FORMATETC& fmtetc)
{
	COleDataObject* pDataObject = pctx->m_pDataObject;
	CString strFormatName;
	BOOL bRegistered = FALSE;
	GetClipboardFormatNameString(fmtetc.cfFormat, strFormatName, bRegistered);

	if (bRegistered)
	{
		// `n[hR[h́BShellnhtml format
		if (CompareNoCaseX(strFormatName, L"CFSTR_SHELLIDLIST") == 0)
		{
			shared_ptr<CCBData> pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			pData->m_suitabletype = suitabletypeProperty;
			pData->m_nIconType = icontypeFolder;
			pData->ParseShellIDList(pctx);
			return pData;
		}
		else if (CompareNoCaseX(strFormatName, L"CFSTR_FILEDESCRIPTORW") == 0)
		{
			shared_ptr<CCBData> pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			shared_ptr<CCBData_FileDescriptor> pData1 = CCBData_FileDescriptor::ChangeClassFrom(pData);
			pData1->ParseFileDescriptorW();
			pData1->m_suitabletype = suitabletypeProperty;
			return pData1;
		}
		else if (CompareNoCaseX(strFormatName, L"CFSTR_FILEDESCRIPTORA") == 0)
		{
			shared_ptr<CCBData> pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			shared_ptr<CCBData_FileDescriptor> pData1 = CCBData_FileDescriptor::ChangeClassFrom(pData);
			pData1->ParseFileDescriptorA();
			pData1->m_suitabletype = suitabletypeProperty;
			return pData1;
		}
		else if (CompareNoCaseX(strFormatName, L"CFSTR_FILECONTENTS") == 0)
		{
			shared_ptr<CCBData> pData = make_shared<CCBData>();
			pData->m_fmtetc = fmtetc;
			pData->m_nIconType = icontypeFolder;
			// g͍ŌFILEDESCRIPTORăZbgB
			return pData;
		}
		else if (CompareNoCaseX(strFormatName, L"CFSTR_FILENAMEMAP") == 0)
		{
			shared_ptr<CCBData> pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			pData->ParseHDrop(pctx, FALSE);
			pData->m_suitabletype = suitabletypeProperty;
			return pData;
		}
		else if (CompareNoCaseX(strFormatName, L"\"DragImageBits\"") == 0)
		{
			shared_ptr<CCBData> pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			shared_ptr<CCBData_DragImageBits> pData1 = CCBData_DragImageBits::ChangeClassFrom(pData);
			return pData1;
		}
		else
		{
			shared_ptr<CCBData> pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			pData->m_suitabletype = suitabletypeHex;
			return pData;
		}

	}
	else
	{
		shared_ptr<CCBData> pData;
		switch (fmtetc.cfFormat)
		{
		case CF_BITMAP:
			pData = CreateSub(pctx, fmtetc, TYMED_GDI);
			return pData;
		case CF_DIB:
		case CF_DIBV5:
			{
				pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
				shared_ptr<CCBData_DIB> pData1 = CCBData_DIB::ChangeClassFrom(pData);
				return pData1;
			}
		case CF_DIF:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			pData->m_suitabletype = suitabletypeText;
			pData->m_nCodePage = 0;
			return pData;
		case CF_DSPBITMAP:
		case CF_DSPENHMETAFILE:
		case CF_DSPMETAFILEPICT:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			return pData;
		case CF_DSPTEXT:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			pData->m_suitabletype = suitabletypeText;
			pData->m_nCodePage = 0;
			return pData;
		case CF_ENHMETAFILE:
			pData = CreateSub(pctx, fmtetc, TYMED_ENHMF);
			return pData;
		case CF_HDROP:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			pData->ParseHDrop(pctx, TRUE);
			pData->m_suitabletype = suitabletypeProperty;
			pData->m_nIconType = icontypeFolder;
			return pData;
		case CF_LOCALE:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			return pData;
		case CF_METAFILEPICT:
			pData = CreateSub(pctx, fmtetc, TYMED_MFPICT);
			return pData;
		case CF_OEMTEXT:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			pData->m_suitabletype = suitabletypeText;
			pData->m_nCodePage = CP_OEMCP;
			return pData;
		case CF_OWNERDISPLAY:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			return pData;
		case CF_PALETTE:
			pData = make_shared<CCBData>();					
			pData->m_fmtetc = fmtetc;
			return pData;
		case CF_PENDATA:
			pData = make_shared<CCBData>();					
			pData->m_fmtetc = fmtetc;
			return pData;
		case CF_RIFF:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			return pData;
		case CF_SYLK:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			pData->m_suitabletype = suitabletypeText;
			pData->m_nCodePage = 0;
			return pData;
		case CF_TEXT:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			pData->m_suitabletype = suitabletypeText;
			pData->m_nCodePage = 0;
			return pData;
		case CF_TIFF:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			return pData;
		case CF_UNICODETEXT:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			pData->m_suitabletype = suitabletypeText;
			pData->m_nCodePage = CP_CBDATA_UTF16;
			return pData;
		case CF_WAVE:
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			return pData;
		}
		if (fmtetc.cfFormat >= CF_GDIOBJFIRST && fmtetc.cfFormat <= CF_GDIOBJLAST)
		{
			pData = make_shared<CCBData>();					
			pData->m_fmtetc = fmtetc;
			return pData;			
		}
		if (fmtetc.cfFormat >= CF_PRIVATEFIRST && fmtetc.cfFormat <= CF_PRIVATELAST)
		{
			pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
			return pData;
		}

		// B
		pData = CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
		return pData;
	}
	return nullptr;
}

shared_ptr<CCBData> CCBData::CreateSub(CCBDataCreateContext* pctx, FORMATETC& fmtetc0, DWORD tymed)
{
/* TYMED_HGLOBAL    = 1,
  TYMED_FILE       = 2,
  TYMED_ISTREAM    = 4,
  TYMED_ISTORAGE   = 8,
  TYMED_GDI        = 16,
  TYMED_MFPICT     = 32,
  TYMED_ENHMF      = 64,_*/

	COleDataObject* pDataObject = pctx->m_pDataObject;
	shared_ptr<CCBData> pData;

	ASSERT(pDataObject);

	STGMEDIUM stgmed = {0};
	BOOL bGetDataOK = FALSE;
	if (pDataObject->GetData(fmtetc0.cfFormat, &stgmed, &fmtetc0))
	{
		bGetDataOK = TRUE;
	}
	else
	{
		// fmtetctymed͐MȂB
		// S
		FORMATETC fmtetc = fmtetc0;
		fmtetc.tymed = TYMED_HGLOBAL | TYMED_FILE | TYMED_ISTREAM | TYMED_ISTORAGE | TYMED_GDI | TYMED_MFPICT | TYMED_ENHMF;
		if (pDataObject->GetData(fmtetc.cfFormat, &stgmed, &fmtetc))
		{
			bGetDataOK = TRUE;
		}
	}
	if (bGetDataOK && stgmed.tymed)
	{
		if (stgmed.tymed & TYMED_HGLOBAL)
			pData = CreateFromHGlobal(pctx, stgmed.hGlobal);
		else if (stgmed.tymed & TYMED_FILE)
			pData = CreateFromFile(pctx, stgmed.lpszFileName);
		else if (stgmed.tymed & TYMED_ISTREAM)
			pData = CreateFromIStream(pctx, stgmed.pstm);
		else if (stgmed.tymed & TYMED_ISTORAGE)
			pData = CreateFromIStorage(pctx, stgmed.pstg);
		else if (stgmed.tymed & TYMED_GDI)
			pData = CreateFromHBITMAP(stgmed.hBitmap, pctx->m_hPalette);
		else if (stgmed.tymed & TYMED_MFPICT)
			pData = CreateFromMFPICT(stgmed.hMetaFilePict);
		else if (stgmed.tymed & TYMED_ENHMF)
			pData = CreateFromHENHMETAFILE(stgmed.hEnhMetaFile);
		else
		{
			pData = CreateErrorData(L"ΉĂȂtymedłB");
		}
		if (pData)
			pData->m_tymed = stgmed.tymed;
			
		ReleaseStgMedium(&stgmed);
	}
	else
	{
		pData = CreateErrorData(L"f[^擾ł܂iCOleDataObject::GetData()j");
	}

	if (!pData)
	{
		pData = CreateErrorData(L"f[^擾ł܂B");
	}
	pData->m_fmtetc = fmtetc0;
	return pData;
}

shared_ptr<CCBData> CCBData::CreateFromHGlobal(CCBDataCreateContext* pctx, HGLOBAL hGlobal)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	shared_ptr<CCBData> pData = make_shared<CCBData>();
	if (!hGlobal)
	{
		pData->m_arrayProperties.Add(L"G[\tf[^擾ł܂iHGLOBAL=NULLjB");
		return pData;
	}

	CGlobalPtr<BYTE*> pGlobalData(hGlobal);
	if (pGlobalData)
	{
		pData->m_cbMem = GlobalSize(hGlobal);
		pData->m_strMemDataFileName = pApp->MakeTempPathName();
		NFile file;
		if (file.CreateNewFileToWrite(pData->m_strMemDataFileName))
		{
			size_t cbRemain = (size_t)pData->m_cbMem;
			size_t nOffset = 0;
			while (cbRemain)
			{
				if (pctx->IsCancelled())
					break;

				UINT cbToWrite = (UINT)min(cbRemain, 1024 * 1024);
				if (!file.Write(pGlobalData + nOffset, cbToWrite))
					break;

				cbRemain -= cbToWrite;
				nOffset += cbToWrite;
			}
			file.Close();
			if (cbRemain)
			{
				// Ȃ
				DeleteFile(pData->m_strMemDataFileName);
				pData->m_strMemDataFileName.Empty();
				pData->m_arrayProperties.Add(L"G[\tꎞt@Cɏ߂܂B");
			}
		}
		else
		{
			pData->m_arrayProperties.Add(L"G[\tꎞt@Cɏ߂܂B");
		}
	}
	else
	{
		pData->m_arrayProperties.Add(L"G[\tf[^擾ł܂iGlobalLockjB");
	}
	return pData;
}

shared_ptr<CCBData> CCBData::CreateFromFile(CCBDataCreateContext* pctx, const WCHAR* pszPathName)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	shared_ptr<CCBData> pData = make_shared<CCBData>();
	pData->m_strType = L"fBXNt@C";
	BOOL bSuccess = FALSE;

	if (pszPathName)
	{
		NFile fileIn;
		if (fileIn.OpenExistingFileToRead(pszPathName))
		{
			pData->m_cbMem = fileIn.GetLength();
			pData->m_strMemDataFileName = pApp->MakeTempPathName();

			NFile fileOut;
			if (fileOut.CreateNewFileToWrite(pData->m_strMemDataFileName))
			{
				UINT cbBuf = 1024 * 1024;
				vector<BYTE> buf(cbBuf);
				BYTE* pBuf = buf.data();
				ULONGLONG cbRemain = pData->m_cbMem;
				while (cbRemain)
				{
					if (pctx->IsCancelled())
						break;
					UINT cbToRead = (ULONG)min(cbBuf, cbRemain);
					UINT cbRead = fileIn.Read(pBuf, cbToRead);
					if (cbRead != cbToRead)
						break;

					fileOut.Write(pBuf, cbToRead);
					cbRemain -= cbToRead;
				}
				fileOut.Close();

				if (cbRemain)
				{
					// Ȃ
					DeleteFile(pData->m_strMemDataFileName);
					pData->m_strMemDataFileName.Empty();
					pData->m_arrayProperties.Add(L"G[\tꎞt@Cɏ߂܂łB");
				}
			}
			fileIn.Close();
		}
		else
		{
			pData->m_arrayProperties.Add(L"G[\tt@C荞߂܂B");
		}
	}
	else
	{
		pData->m_arrayProperties.Add(L"G[\tt@C荞߂܂B");
	}
	return pData;
}

shared_ptr<CCBData> CCBData::CreateFromIStream(CCBDataCreateContext* pctx, IStream* pstm)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	shared_ptr<CCBData> pData = make_shared<CCBData>();
	pData->m_strType = L"IStream";

	if (!pstm)
	{
		pData->m_arrayProperties.Add(L"G[\tf[^擾ł܂iIStream*NULLjB");
		return pData;
	}

	STATSTG statstg;
	if (SUCCEEDED(pstm->Stat(&statstg, STATFLAG_NONAME)))
	{
		LARGE_INTEGER liZero;
		liZero.QuadPart = 0;
		ULARGE_INTEGER liPosition;
		pstm->Seek(liZero, STREAM_SEEK_SET, &liPosition);

		pData->m_cbMem = statstg.cbSize.QuadPart;
		pData->m_strMemDataFileName = pApp->MakeTempPathName();
		NFile file;
		if (file.CreateNewFileToWrite(pData->m_strMemDataFileName))
		{
			ULONGLONG cbRemain = pData->m_cbMem;
			static const int cbBuf = 1024 * 1024;
			vector<BYTE> buf(cbBuf);
			BYTE* pBuf = buf.data();
			while (cbRemain)
			{
				if (pctx->IsCancelled())
				{
					break;
				}

				ULONG cbToRead = (ULONG)min(cbBuf, cbRemain);
				ULONG cbRead;
				if (!SUCCEEDED(pstm->Read(pBuf, cbToRead, &cbRead)))
					break;
				if (cbRead != cbToRead)
					break;

				if (!file.Write(pBuf, cbBuf)) 
					break;

				cbRemain -= cbToRead;
			}
			file.Close();

			if (cbRemain)
			{
				DeleteFile(pData->m_strMemDataFileName);
				pData->m_strMemDataFileName.Empty();
				pData->m_arrayProperties.Add(L"G[\tXg[ꎞt@CɃRs[ł܂B");
			}
		}
		else
		{
			pData->m_arrayProperties.Add(L"G[\tꎞt@Cɏ߂܂B");
		}
	}
	else
	{
		pData->m_arrayProperties.Add(L"G[\tXg[̃TCY擾ł܂B");
	}
	return pData;
}

shared_ptr<CCBData> CCBData::CreateFromIStorage(CCBDataCreateContext* pctx, IStorage* pStorage)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	shared_ptr<CCBData> pData = make_shared<CCBData>();
	pData->m_strType = L"IStorage";
	pData->m_nIconType = icontypeFolder;

	if (!pStorage)
	{
		pData->m_arrayProperties.Add(L"G[\tf[^擾ł܂iIStorage*NULLjB");
		return pData;
	}

	pData->m_pArrayChildren = make_shared<CCBDataArray>();
	pData->m_pArrayChildren->CreateSubTreeFromIStorage(pctx, pStorage);
	return pData;
}

CString GetResultAndFree(STRRET* pstrret)
{
	CString str;
	if (pstrret->uType == STRRET_WSTR)
	{
		str = pstrret->pOleStr;
		CoTaskMemFree(pstrret->pOleStr);
	}
	else if (pstrret->uType == STRRET_CSTR)
	{
		str = pstrret->cStr;
		// ɂf[^Ȃ̂ŉ͂Ȃ
	}
	return str;
}

shared_ptr<CCBData> CCBData::CreateFromIShellFolder(CCBDataCreateContext* pctx, IShellFolder* pFolder)
{
	shared_ptr<CCBData> pData = make_shared<CCBData>();
	pData->m_strType = L"IShellFolder";
	pData->m_nIconType = icontypeFolder;

	if (!pFolder)
	{
		pData->m_arrayProperties.Add(L"G[\tf[^擾ł܂(IShellFolder*NULL)B");
		return pData;
	}

	pData->m_pArrayChildren = make_shared<CCBDataArray>();
	pData->m_pArrayChildren->CreateSubTreeFromIShellFolder(pctx, pFolder);

	return pData;
}

shared_ptr<CCBData> CCBData::CreateFromDirectory(CCBDataCreateContext* pctx, const CString& strDir)
{
	shared_ptr<CCBData> pData = make_shared<CCBData>();
	pData->m_strType = L"fBNg";
	pData->m_nIconType = icontypeFolder;

	pData->m_pArrayChildren = make_shared<CCBDataArray>();
	pData->m_pArrayChildren->CreateSubTreeFromDirectory(pctx, strDir);
	return pData;
}

shared_ptr<CCBData> CCBData::CreateFromHBITMAP(HBITMAP hBitmap, HPALETTE hPalette)
{
	shared_ptr<CCBData_HBITMAP> pData = make_shared<CCBData_HBITMAP>();
	pData->CopyImage(hBitmap, hPalette);

	return pData;
}

shared_ptr<CCBData> CCBData::CreateFromMFPICT(HMETAFILEPICT hMetaFilePict)
{
	shared_ptr<CCBData_HMETAFILEPICT> pData = make_shared<CCBData_HMETAFILEPICT>();
	pData->CopyImage(hMetaFilePict);

	return pData;
}

shared_ptr<CCBData> CCBData::CreateFromHENHMETAFILE(HENHMETAFILE hEnhMetaFile)
{
	shared_ptr<CCBData_HENHMF> pData = make_shared<CCBData_HENHMF>();
	pData->CopyImage(hEnhMetaFile);
	return pData;
}

shared_ptr<CCBData> CCBData::CreateDummyFolder(const CString& strType)
{
	shared_ptr<CCBData> pData = make_shared<CCBData>();
	pData->m_bPlaceHolder = TRUE;
	pData->m_strType = strType;
	pData->m_nIconType = icontypeFolder;
	return pData;
}

shared_ptr<CCBData> CCBData::CreateErrorData(const CString& strError)
{
	shared_ptr<CCBData> pData = make_shared<CCBData>();
	pData->m_arrayProperties.Add(CString(L"G[\t") + strError);
	return pData;
}

BOOL CCBData::ParseHDrop(CCBDataCreateContext* pctx, BOOL bAddFiles)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	vector<BYTE> buf;
	if (!LoadMemData(buf))
		return FALSE;

	HDROP hDrop = (HDROP)buf.data();
	if (bAddFiles)
	{
		ASSERT(m_pArrayChildren == NULL);
		m_pArrayChildren = make_shared<CCBDataArray>();
		m_pArrayChildren->CreateSubTreeFromHDROP(pctx, hDrop);
	}
	else
	{
		CString strCol0 = L"t@Cꗗ\t";

		UINT nCount = DragQueryFileCount(hDrop);
		for (UINT i = 0; i < nCount; i++)
		{
			CString strFileName;
			if (DragQueryFile(hDrop, i, strFileName))
				m_arrayProperties.Add(strCol0 + strFileName);
			else
				m_arrayProperties.Add(strCol0 + L"(擾ł܂)");

			strCol0 = L"\t";
		}
	}
	return TRUE;
}

#define GetPIDLFolder(pida) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[0])
#define GetPIDLItem(pida, i) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[i+1])

BOOL CCBData::ParseShellIDList(CCBDataCreateContext* pctx)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	vector<BYTE> buf;
	if (!LoadMemData(buf))
		return FALSE;

	CIDA* pida = (CIDA*)buf.data();
	if (pida->cidl)
	{
		// e擾
		LPCITEMIDLIST pidl = GetPIDLFolder(pida);
		CComPtr<IShellFolder> pFolderDesktop;
		if (SUCCEEDED(SHGetDesktopFolder(&pFolderDesktop)))
		{
			CString strParentName;
			STRRET strretName;
			if (SUCCEEDED(pFolderDesktop->GetDisplayNameOf(pidl, SHGDN_NORMAL, &strretName)))
			{
				strParentName = GetResultAndFree(&strretName);
			}
			if (SUCCEEDED(pFolderDesktop->GetDisplayNameOf(pidl, SHGDN_FORPARSING, &strretName)))
			{
				CString str = GetResultAndFree(&strretName);
				strParentName += L" (";
				strParentName += str;
				strParentName += L")";
			}
			m_arrayProperties.Add(CString(L"etH_\t") + strParentName);

			CComPtr<IShellFolder> pFolderParent;
			if (pidl->mkid.cb == 0)
			{
				pFolderParent = pFolderDesktop;
			}
			else
			{
				if (SUCCEEDED(pFolderDesktop->BindToObject(pidl, NULL, IID_IShellFolder, (void**)&pFolderParent)))
				{
				}
			}
			if (pFolderParent)
			{
				ASSERT(m_pArrayChildren == NULL);
				m_pArrayChildren = make_shared<CCBDataArray>();
				m_pArrayChildren->CreateSubTreeFromCIDA(pctx, pFolderParent, pida);
			}
		}
	}
	return TRUE;
}

shared_ptr<Image> CCBData::GetImage(int& cx, int& cy, int& nBitsPerPixel)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	cx = 0;
	cy = 0;
	nBitsPerPixel = 0;

	BOOL b = FALSE;
	if (m_cbMem)
	{
		if (!m_strMemDataFileName.IsEmpty())
		{
			shared_ptr<Bitmap> pBitmap;
			bool bSuccess = false;
			if (pApp->m_pWICFactory) {
				CComPtr<IWICBitmapDecoder> pBitmapDecoder;
				if (SUCCEEDED(pApp->m_pWICFactory->CreateDecoderFromFilename(m_strMemDataFileName, nullptr, GENERIC_READ, WICDecodeMetadataCacheOnLoad, &pBitmapDecoder))) {
					CComPtr<IWICBitmapFrameDecode> pFrameDecode;
					if (SUCCEEDED(pBitmapDecoder->GetFrame(0, &pFrameDecode))) {
						// orientation
						int nOrientation = 0;		// N/A
						CComPtr<IWICMetadataQueryReader> pMetadataReader;
						if (SUCCEEDED(pFrameDecode->GetMetadataQueryReader(&pMetadataReader))) {
							PROPVARIANT v;
							PropVariantInit(&v);
							if (SUCCEEDED(pMetadataReader->GetMetadataByName(L"/app1/ifd/{ushort=274}", &v))) {
								if (v.vt == VT_UI2) {
									nOrientation = v.uiVal;
								}
							}
							else {
								PropVariantInit(&v);
								if (SUCCEEDED(pMetadataReader->GetMetadataByName(L"/ifd/{ushort=274}", &v))) {
									if (v.vt == VT_UI2) {
										nOrientation = v.uiVal;
									}
								}
							}
						}

						WICPixelFormatGUID guidPixelFormat;
						if (SUCCEEDED(pFrameDecode->GetPixelFormat(&guidPixelFormat))) {
							CComPtr<IWICComponentInfo> pComponentInfo;
							if (SUCCEEDED(pApp->m_pWICFactory->CreateComponentInfo(guidPixelFormat, &pComponentInfo))) {
								CComPtr<IWICPixelFormatInfo> pPixelFormatInfo;
								if (SUCCEEDED(pComponentInfo->QueryInterface(__uuidof(IWICPixelFormatInfo), (void**)&pPixelFormatInfo))) {
									UINT uBPP;
									if (SUCCEEDED(pPixelFormatInfo->GetBitsPerPixel(&uBPP))) {
										nBitsPerPixel = (int)uBPP;
									}
								}
							}
						}


						CComPtr<IWICFormatConverter> pFormatConverter;
						if (SUCCEEDED(pApp->m_pWICFactory->CreateFormatConverter(&pFormatConverter))) {
							if (SUCCEEDED(pFormatConverter->Initialize(pFrameDecode, GUID_WICPixelFormat32bppBGRA, WICBitmapDitherTypeNone, nullptr, 0.0f, WICBitmapPaletteTypeMedianCut))) {
								UINT width, height;
								if (SUCCEEDED(pFormatConverter->GetSize(&width, &height))) {
									pBitmap = make_shared<Bitmap>(width, height, PixelFormat32bppARGB);
									Rect rc(0, 0, width, height);
									BitmapData bmd;
									if (pBitmap->LockBits(&rc, ImageLockModeWrite, PixelFormat32bppARGB, &bmd) == Ok) {
										if (bmd.Stride < 0) {
											// tH
										}
										else {
											if (SUCCEEDED(pFormatConverter->CopyPixels(NULL, bmd.Stride, bmd.Stride * height, (BYTE*)bmd.Scan0))) {
												bSuccess = true;
											}
										}
										pBitmap->UnlockBits(&bmd);
										if (nOrientation) {
											RotateExif(pBitmap.get(), nOrientation);
										}
										cx = pBitmap->GetWidth();
										cy = pBitmap->GetHeight();
									}
								}
							}
						}
					}
				}
			}
			if (bSuccess) {
				return pBitmap;
			}
		}
	}
	return nullptr;
}

CString CCBData::GetTitle()
{
	if (!m_strDispFileName.IsEmpty())
	{
		return m_strDispFileName;
	}
	else if (!m_strFileName.IsEmpty())
	{
		return m_strFileName;
	}
	else
	{
		return FORMATETC2ShortString(m_fmtetc);
	}
}

map<CompareXString, CString> g_mapFormatsDefaultViewTypeTable =
{

	{	L"\"HTML Format\"", L"TEXT"	},
	{	L"CFSTR_FILENAMEA",	L"TEXT"	},
	{	L"CFSTR_FILENAMEW",	L"TEXT"	},
	{	L"CFSTR_INETURLA",	L"TEXT"	},
	{	L"CFSTR_INETURLW",	L"TEXT"	},
	{	L"CFSTR_SHELLURL",	L"TEXT"	},

	{	L".png",			L"IMAGE"	},
	{	L".jpg",			L"IMAGE"	},
	{	L".jpeg",			L"IMAGE"	},
	{	L".gif",			L"IMAGE"	},
	{	L".bmp",			L"IMAGE"	},
	{	L".tiff",			L"IMAGE"	},
	{	L".exif",			L"IMAGE"	},
	{	L".jfif",			L"IMAGE"	},
	{	L".heif",			L"IMAGE"	},
	{	L".heic",			L"IMAGE"	},

	{	L".url",			L"TEXT"		},

	{	L"\"application/microdata+json\"",			L"TEXT"	},
	{	L"\"application/x-moz-file-promise-url\"",	L"TEXT"	},
	{	L"\"application/x-moz-file-promise-dest-filename\"",	L"TEXT"	},
	{	L"\"Csv\"",									L"TEXT"	},
	{	L"\"msSourceUrl\"",							L"TEXT"	},
	{	L"\"GIF\"",									L"IMAGE"	},
	{	L"\"JFIF\"",								L"IMAGE"	},
	{	L"\"PNG\"",									L"IMAGE"	},
	{	L"\"Rich Text Format\"",					L"TEXT"	},
	{	L"\"XML Spreadsheet\"",						L"TEXT"	},
};


map<CString, UINT> g_mapFormatsDefaultViewCodePageTable =
{
	{	L"\"HTML Format\"", (UINT)CP_UTF8	},
	{	L"CFSTR_FILENAMEA",	(UINT)CP_ACP	},
	{	L"CFSTR_FILENAMEW",	(UINT)CP_CBDATA_UTF16	},
	{	L"CFSTR_INETURLA",	(UINT)CP_ACP	},
	{	L"CFSTR_INETURLW",	(UINT)CP_CBDATA_UTF16	},
	{	L"CFSTR_SHELLURL",	(UINT)CP_ACP	},
};


int CCBData::GetSuitableType()
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	CompareXString strFormat = GetFormatTypeString();

	// WXg
	CString strType = pApp->GetProfileString(L"FormatsDefaultViewTab", strFormat, L"");
	if (strType.IsEmpty())
	{
		// e[u
		auto iter = g_mapFormatsDefaultViewTypeTable.find(strFormat);
		if (iter != g_mapFormatsDefaultViewTypeTable.end()) 
		{
			strType = iter->second;
		}
	}
	// "text/*"^Cv̓eLXg낤B
	if (strType.IsEmpty())
	{
		if (CompareNoCaseX(strFormat.Left(6), L"\"text/") == 0)
		{
			strType = L"TEXT";
		}
	}

	if (!strType.IsEmpty())
	{
		if (CompareNoCaseX(strType, L"PROPERTY") == 0)
			return suitabletypeProperty;
		else if (CompareNoCaseX(strType, L"HEX") == 0)
			return suitabletypeHex;
		else if (CompareNoCaseX(strType, L"TEXT") == 0)
			return suitabletypeText;
		else if (CompareNoCaseX(strType, L"IMAGE") == 0)
			return suitabletypeImage;
	}
	
	return m_suitabletype;
}

void CCBData::SetSuitableType(int suitabletype)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	CString strFormat = GetFormatTypeString();

	CString strType;
	switch (suitabletype)
	{
	case suitabletypeHex:
		strType = L"HEX";
		break;
	case suitabletypeText:
		strType = L"TEXT";
		break;
	case suitabletypeImage:
		strType = L"IMAGE";
		break;
	case suitabletypeProperty:
		strType = L"PROPERTY";
		break;
	default:
		return;
	}

	pApp->WriteProfileString(L"FormatsDefaultViewTab", strFormat, strType);
}

int CCBData::GetSuitableCodePage()
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	CString strFormat = GetFormatTypeString();

	// WXg
	// WXg͎Ȃقg₷悤
	static const int CP_UNDEFINED = -100;
	int nCodePage = CP_UNDEFINED;
//	int nCodePage = pApp->GetProfileInt(L"FormatsDefaultViewCodePage", strFormat, CP_UNDEFINED);
//	if (nCodePage == CP_UNDEFINED)
	{
		// e[u
		auto iter = g_mapFormatsDefaultViewCodePageTable.find(strFormat);
		if (iter != g_mapFormatsDefaultViewCodePageTable.end())
		{
			nCodePage = iter->second;
		}
	}
	if (nCodePage != CP_UNDEFINED)
		return nCodePage;
	
	return m_nCodePage;
}

//void CCBData::SetSuitableCodePage(int nSuitableCodePage)
//{
//	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

//	CString strFormat = GetFormatTypeString();

//	CString strType;
//	pApp->WriteProfileInt(L"FormatsDefaultViewCodePage", strFormat, nSuitableCodePage);
//}


CString CCBData::GetFormatTypeString()
{
	// t@CȂgq
	CString strFormat;
	if (!m_strFileName.IsEmpty())
	{
		const WCHAR* pchStart = (const WCHAR*)m_strFileName;
		const WCHAR* pch = pchStart + m_strFileName.GetLength() - 1;

		while (pch >= pchStart)
		{
			if (*pch == L'.')
			{
				strFormat = pch;
				break;
			}
			pch--;
		}
	}
	if (strFormat.IsEmpty())
	{
		strFormat = GetClipboardFormatNameString(m_fmtetc.cfFormat);
	}
	return strFormat;
}

/////////////////////////////////////

IMPLEMENT_DYNAMIC(CCBData_FileDescriptor, CCBData)

CCBData_FileDescriptor::CCBData_FileDescriptor()
{
}

CCBData_FileDescriptor::~CCBData_FileDescriptor()
{
}

BOOL CCBData_FileDescriptor::ParseFileDescriptorW()
{
	vector<BYTE> buf;
	if (LoadMemData(buf))
	{
		FILEGROUPDESCRIPTORW* pfgd = (FILEGROUPDESCRIPTORW*)buf.data();
		for (UINT i = 0; i < pfgd->cItems; i++)
		{
			CString strFileName(pfgd->fgd[i].cFileName);
			m_arrayProperties.Add(CString(L"O[") + UnsignedDecimalString(i) + L"]\t" + strFileName);
			m_arrayFileNames.Add(strFileName);
		}
	}
	return TRUE;
}

BOOL CCBData_FileDescriptor::ParseFileDescriptorA()
{
	vector<BYTE> buf;
	if (LoadMemData(buf))
	{
		FILEGROUPDESCRIPTORA* pfgd = (FILEGROUPDESCRIPTORA*)buf.data();
		for (UINT i = 0; i < pfgd->cItems; i++)
		{
			CString strFileName(pfgd->fgd[i].cFileName);
			m_arrayProperties.Add(CString(L"O[") + UnsignedDecimalString(i) + L"]\t" + strFileName);
			m_arrayFileNames.Add(strFileName);
		}
	}
	return TRUE;
}

/////////////////////////////////////

IMPLEMENT_DYNAMIC(CCBData_GDI, CCBData)

shared_ptr<Image> CCBData_GDI::GetImage(int& cx, int& cy, int& nBitsPerPixel)
{
	shared_ptr<Image> pImage = GetImageSub();

	if (m_cxImage)
	{
		cx = m_cxImage;
		cy = m_cyImage;
		nBitsPerPixel = m_nBitsPerPixel;
	}
	else if (pImage)
	{
		cx = pImage->GetWidth();
		cy = pImage->GetHeight();
		nBitsPerPixel = 0;
	}
	else
	{
		cx = cy = nBitsPerPixel = 0;
	}

	return pImage;
}


/////////////////////////////////////

IMPLEMENT_DYNAMIC(CCBData_HBITMAP, CCBData_GDI)

CCBData_HBITMAP::CCBData_HBITMAP()
: m_hBitmap(NULL)
, m_hPalette(NULL)
{
}

CCBData_HBITMAP::~CCBData_HBITMAP()
{
	if (m_hBitmap)
		DeleteObject(m_hBitmap);
	if (m_hPalette)
		DeleteObject(m_hPalette);
}

void CCBData_HBITMAP::CopyImage(HBITMAP hBitmap, HPALETTE hPal)
{
	ASSERT(m_hBitmap == NULL);
	ASSERT(m_hPalette == NULL);

	if (hBitmap)
		m_hBitmap = (HBITMAP)OleDuplicateData(hBitmap, CF_BITMAP, 0);
	if (hPal)
		m_hPalette = (HPALETTE)OleDuplicateData(hPal, CF_PALETTE, 0);

	if (hBitmap)
	{
		BITMAP bm = {0};
		GetObject(hBitmap, sizeof(bm), &bm);
		m_cxImage = bm.bmWidth;
		m_cyImage = bm.bmHeight;
		m_nBitsPerPixel = bm.bmBitsPixel * bm.bmPlanes;
	}
}

shared_ptr<Image> CCBData_HBITMAP::GetImageSub()
{
	if (!m_hBitmap)
		return nullptr;

	return shared_ptr<Image>(Bitmap::FromHBITMAP(m_hBitmap, m_hPalette));
}

/////////////////////////////////////

IMPLEMENT_DYNAMIC(CCBData_HENHMF, CCBData_GDI)

CCBData_HENHMF::CCBData_HENHMF()
: m_hEM(NULL)
{
}

CCBData_HENHMF::~CCBData_HENHMF()
{
	if (m_hEM)
		DeleteObject(m_hEM);
}

void CCBData_HENHMF::CopyImage(HENHMETAFILE hEM)
{
	ASSERT(m_hEM == NULL);

	if (hEM)
		m_hEM = (HENHMETAFILE)OleDuplicateData(hEM, CF_ENHMETAFILE, 0);
}

shared_ptr<Image> CCBData_HENHMF::GetImageSub()
{
	if (!m_hEM)
		return NULL;
	return make_shared<Metafile>(m_hEM);
}

/////////////////////////////////////

IMPLEMENT_DYNAMIC(CCBData_HMETAFILEPICT, CCBData_GDI)

CCBData_HMETAFILEPICT::CCBData_HMETAFILEPICT()
: m_hMetaFilePict(NULL)
, m_hMF(NULL)
{
}

CCBData_HMETAFILEPICT::~CCBData_HMETAFILEPICT()
{
	if (m_hMF)
		DeleteObject(m_hMF);

	if (m_hMetaFilePict)
		GlobalFree(m_hMetaFilePict);
}

void CCBData_HMETAFILEPICT::CopyImage(HMETAFILEPICT hMetaFilePict)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	// 1񂵂ĂׂȂ
	ASSERT(m_hMetaFilePict == NULL);

	if (hMetaFilePict)
		m_hMetaFilePict = (HMETAFILEPICT)OleDuplicateData(hMetaFilePict, CF_METAFILEPICT, 0);

	{
		CGlobalPtr<METAFILEPICT*> pmf(m_hMetaFilePict);
		if (pmf && pmf->hMF)
		{
			m_hMF = pmf->hMF;
			// ĂBTCY͂͂
			m_cbMem = GlobalSize(m_hMetaFilePict);
			m_strMemDataFileName = pApp->MakeTempPathName();
			NFile file;
			if (file.CreateNewFileToWrite(m_strMemDataFileName))
			{
				file.Write(pmf, (UINT)m_cbMem);
			}
		}
	}
}

shared_ptr<Image> CCBData_HMETAFILEPICT::GetImageSub()
{
	if (!m_hMF)
		return nullptr;
#ifdef _DEBUG
#undef new
#endif
	return shared_ptr<Metafile>(new Metafile(m_hMF, NULL));
#ifdef _DEBUG
#define new DEBUG_NEW
#endif
}

/////////////////////////////////////

IMPLEMENT_DYNAMIC(CCBData_DIB, CCBData)

CCBData_DIB::CCBData_DIB()
{
	m_suitabletype = suitabletypeImage;
}

CCBData_DIB::~CCBData_DIB()
{
}

shared_ptr<Image> CCBData_DIB::GetImage(int& cx, int& cy, int& nBitsPerPixel)
{
	if (m_buf.empty()) {
		LoadMemData(m_buf);
	}

	if (m_buf.size() >= sizeof(BITMAPINFOHEADER))
	{
		BITMAPINFO* pbi = (BITMAPINFO*)m_buf.data();

		cx = pbi->bmiHeader.biWidth;
		cy = pbi->bmiHeader.biHeight;
		nBitsPerPixel = pbi->bmiHeader.biBitCount * pbi->bmiHeader.biPlanes;

		// J[e[ũTCY߂
		int nColors = 0;
		switch (pbi->bmiHeader.biBitCount)
		{
		case 0:
			nColors = 0;
			break;
		case 1:
			nColors = 2;
			break;
		case 4:
			if (pbi->bmiHeader.biClrUsed)
				nColors = pbi->bmiHeader.biClrUsed;
			else
				nColors = 16;
			break;
		case 8:
			if (pbi->bmiHeader.biClrUsed)
				nColors = pbi->bmiHeader.biClrUsed;
			else
				nColors = 256;
			break;
		case 16:
			if (pbi->bmiHeader.biCompression == BI_RGB)
				nColors = pbi->bmiHeader.biClrUsed;
			else if (pbi->bmiHeader.biCompression == BI_BITFIELDS)
				nColors = 3 + pbi->bmiHeader.biClrUsed;
			break;
		case 24:
			nColors = pbi->bmiHeader.biClrUsed;
			break;
		case 32:
			if (pbi->bmiHeader.biCompression == BI_RGB)
				nColors = pbi->bmiHeader.biClrUsed;
			else if (pbi->bmiHeader.biCompression == BI_BITFIELDS)
				nColors = 3 + pbi->bmiHeader.biClrUsed;
			break;
		}

		BYTE* pColor = (BYTE*)((LPSTR)pbi + (WORD)(pbi->bmiHeader.biSize));
		BYTE* pPixels = pColor + (nColors * 4);

		return shared_ptr<Image>(Bitmap::FromBITMAPINFO(pbi, pPixels));
	}
	cx = cy = nBitsPerPixel = 0;
	return nullptr;
}

/////////////////////////////////////

IMPLEMENT_DYNAMIC(CCBData_DragImageBits, CCBData)

CCBData_DragImageBits::CCBData_DragImageBits()
{
	m_suitabletype = suitabletypeImage;
}

CCBData_DragImageBits::~CCBData_DragImageBits()
{
}

shared_ptr<Image> CCBData_DragImageBits::GetImage(int& cx, int& cy, int& nBitsPerPixel)
{
	cx = cy = nBitsPerPixel = 0;

	if (m_buf.empty())
		LoadMemData(m_buf);

	if (!m_buf.empty())
	{
		// 32bitł64bitłB
		SHDRAGIMAGE* pdi = (SHDRAGIMAGE*)m_buf.data();

		size_t cb = (size_t)pdi->sizeDragImage.cx * pdi->sizeDragImage.cy * 4;

		int nOffset = 0;
		if (m_cbMem >= cb + 0x0020)
		{
			// Ԃ64bit
			nOffset = 0x0020;
		}
		else if (m_cbMem >= cb + 0x0018)
		{
			// Ԃ32bit
			nOffset = 0x0018;
		}
		else
		{
			return NULL;
		}
		cx = pdi->sizeDragImage.cx;
		cy = pdi->sizeDragImage.cy;
		nBitsPerPixel = 32;

		shared_ptr<Bitmap> pBitmap = 
			make_shared<Bitmap>(pdi->sizeDragImage.cx,
								pdi->sizeDragImage.cy,
								-pdi->sizeDragImage.cx * 4,
								PixelFormat32bppPARGB,
								m_buf.data() + nOffset + cb - pdi->sizeDragImage.cx * 4);
				
		return pBitmap;
	}
	return NULL;

}

////////////////////////////

void CCBDataArray::EnumAllFormats(COleDataObject* pDataObject, vector<FORMATETC>& arrayFMTETC)
{
	ASSERT(pDataObject);

	pDataObject->BeginEnumFormats();

	TRACE("---- COleDataObject -----\n");
	while (TRUE)
	{
		FORMATETC fmtetc = {0};
		if (!pDataObject->GetNextFormat(&fmtetc))
			break;

		TRACE(L"%u\n", fmtetc.cfFormat);

		arrayFMTETC.push_back(fmtetc);
	}

	// FMTETCʂ̃f[^͎Ȃ疳ʂł낤B
	// dΔrĂ
	// GetCanonicalFormatEtcƂĂԂׂH
	for (int i = 0; i < (int)arrayFMTETC.size() - 1; i++)
	{
		for (int j = (int)arrayFMTETC.size() - 1; j >= i + 1; j--)
		{
			FORMATETC& fmtetcI = arrayFMTETC[i];
			FORMATETC& fmtetcJ = arrayFMTETC[j];
			if (fmtetcI.cfFormat == g_cfFileDescriptorW &&
				fmtetcJ.cfFormat == g_cfFileDescriptorW)
			{
				TRACE("\n");
			}


			if (fmtetcI.cfFormat == fmtetcJ.cfFormat &&
				fmtetcI.dwAspect == fmtetcJ.dwAspect &&
				fmtetcI.lindex == fmtetcJ.lindex &&
				fmtetcI.ptd == fmtetcJ.ptd &&
				fmtetcI.tymed == fmtetcJ.tymed)
			{
				arrayFMTETC.erase(arrayFMTETC.begin() + j);
			}
		}
	}
}

void CCBDataArray::CreateItems(COleDataObject* pDataObject, vector<FORMATETC>& arrayFMTETC)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	// 荞ݒ_CAO\
	CCopyingDialog dlg;
	dlg.Create();
	dlg.ShowWindow(SW_SHOW);
	pApp->m_pMainWnd->EnableWindow(FALSE);
	dlg.EnableWindow(TRUE);
	dlg.SetFocus();

	CCBDataCreateContext ctx(pDataObject, &dlg);
	for (int i = 0; i < (int)arrayFMTETC.size(); i++)
	{
		ctx.SetName(arrayFMTETC[i]);
		if (ctx.IsCancelled())
			break;
		ctx.m_cbTotal = 0;
		shared_ptr<CCBData> pData = CCBData::Create(&ctx, arrayFMTETC[i]);
		if (pData)
			push_back(pData);
		else
		{
			shared_ptr<CCBData> pData = CCBData::CreateErrorData(L"荞݂ł܂łB");
			pData->m_fmtetc = arrayFMTETC[i];
			push_back(pData);
		}
	}

	// Ŝ̒
	// CF_TEXT̃ftHgR[hy[WCF_LOCALE擾
	int nCodePageForCfText = 0;
	BOOL bExplicitCodePage = FALSE;
	for (int i = 0; i < (int)size(); i++)
	{
		auto pData = at(i).get();
		if (pData->m_fmtetc.cfFormat == CF_LOCALE)
		{
			DWORD dw;
			vector<BYTE> buf;
			if (pData->LoadMemData(buf))
			{
				if (GetLocaleInfo(*(LCID*)buf.data(), LOCALE_IDEFAULTANSICODEPAGE | LOCALE_RETURN_NUMBER, (LPTSTR)&dw, sizeof(dw)))
				{
					nCodePageForCfText = dw;
					bExplicitCodePage = TRUE;
					break;
				}
			}
		}
	}

	// CF_TEXTɃZbg
	if (bExplicitCodePage)
	{
		for (int i = 0; i < (int)size(); i++)
		{
			auto pData = at(i).get();
			if (pData->m_fmtetc.cfFormat == CF_TEXT)
			{
				pData->m_nCodePage = nCodePageForCfText;
				pData->m_arrayProperties.Add(CString(L"R[hy[W\t") + UnsignedDecimalString(nCodePageForCfText));
			}
		}
	}

	for (int i = 0; i < (int)size(); i++)
	{
		auto pData = at(i).get();
		if (pData->m_fmtetc.cfFormat == g_cfFileContents)
		{
			ctx.SetName(pData->m_fmtetc);
			if (ctx.IsCancelled())
				break;

			// AXyNgFILEDESCRITOR
			CCBData_FileDescriptor* pDataDescriptor = FindFileDescriptor(pData);

			if (pDataDescriptor)
			{
				pData->m_pArrayChildren = make_shared<CCBDataArray>();
				pData->m_pArrayChildren->CreateSubTreeFromFileDescriptor(&ctx, pDataDescriptor, pData->m_fmtetc);
			}
		}
	}

	// 荞ݒ_CAO
	pApp->m_pMainWnd->EnableWindow(TRUE);
	dlg.DestroyWindow();
	pApp->m_pMainWnd->SetFocus();
	if (ctx.IsCancelled())
		AfxMessageBox(L"~܂B");
}

CCBData_FileDescriptor* CCBDataArray::FindFileDescriptor(CCBData* pDataContents)
{
	CCBData* pDataA;
	CCBData* pDataW;
	for (int i = 0; i < (int)size(); i++)
	{
		auto pData = at(i);
		if (pData->m_fmtetc.cfFormat == g_cfFileDescriptorA)
		{
			if (pData->m_fmtetc.dwAspect == pDataContents->m_fmtetc.dwAspect)
			{
				pDataA = pData.get();
			}
		}
		else if (pData->m_fmtetc.cfFormat == g_cfFileDescriptorW)
		{
			if (pData->m_fmtetc.dwAspect == pDataContents->m_fmtetc.dwAspect)
			{
				pDataW = pData.get();
			}
		}
	}
	if (pDataW)
	{
		ASSERT(pDataW->IsKindOf(RUNTIME_CLASS(CCBData_FileDescriptor)));
		return (CCBData_FileDescriptor*)pDataW;
	}
	else if (pDataA)
	{
		ASSERT(pDataA->IsKindOf(RUNTIME_CLASS(CCBData_FileDescriptor)));
		return (CCBData_FileDescriptor*)pDataA;
	}
	return NULL;
}

void CCBDataArray::CreateSubTreeFromCIDA(CCBDataCreateContext* pctx, IShellFolder* pFolder, CIDA* pida)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	ASSERT(empty());	// ܂ȂƂɂȂ

	for (UINT i = 0; i < pida->cidl; i++)
	{
		LPCITEMIDLIST pidl = GetPIDLItem(pida, i);
		STRRET strretFile;
		CString strFileName;
		CString strDispFileName;
		if (SUCCEEDED(pFolder->GetDisplayNameOf(pidl, SHGDN_FORPARSING | SHGDN_INFOLDER, &strretFile)))
		{
			strFileName = GetResultAndFree(&strretFile);
		}
		else
		{
			strFileName = L"(擾ł܂)";
		}
		if (SUCCEEDED(pFolder->GetDisplayNameOf(pidl, SHGDN_NORMAL, &strretFile)))
		{
			strDispFileName = GetResultAndFree(&strretFile);
		}

		pctx->SetName(strFileName);
		if (pctx->IsCancelled())
			break;

		{
			shared_ptr<CCBData> pData;
			CComPtr<IShellFolder> pShellFolder;
			CComPtr<IStream> pStream;
			CComPtr<IStorage> pStorage;
			if (pApp->m_nHowToGetShellIDList == pApp->getshellidlistIStorage && 
				SUCCEEDED(pFolder->BindToStorage(pidl, NULL, IID_IStorage, (void**)&pStorage)))
			{
				// _~[f[^ĂBTCYOKȂ牺Ŏ擾
				pData = CCBData::CreateDummyFolder(L"IStorage");
				pData->m_pStorage = pStorage;
			}
			else if (SUCCEEDED(pFolder->BindToObject(pidl, NULL, IID_IStream, (void**)&pStream)))
			{
				pData = CCBData::CreateFromIStream(pctx, pStream);
				pctx->m_cbTotal += pData->m_cbMem;
			}
			else if (SUCCEEDED(pFolder->BindToObject(pidl, NULL, IID_IShellFolder, (void**)&pShellFolder)))
			{
				// _~[f[^ĂBTCYOKȂ牺Ŏ擾
				pData = CCBData::CreateDummyFolder(L"IShellFolder");
				pData->m_pShellFolder = pShellFolder;
			}
			else
			{
				pData = make_shared<CCBData>();
				pData->m_strType = L"s";
				pData->m_nIconType = CCBData::icontypeUnknown;
			}

			if (pData)
			{
				pData->m_strFileName = strFileName;
				pData->m_strDispFileName = strDispFileName;

				push_back(pData);
			}
		}
	}
	// TCYE𒴂ĂȂ΃TutH_
	ULONGLONG cbLimit = (ULONGLONG)pApp->m_nSubDirSizeLimit * 1024 * 1024;
	for (int i = 0; i < (int)size(); i++)
	{
		CCBData* pData = at(i).get();
		if (pData->m_bPlaceHolder)
		{
			pctx->SetName(pData->m_strFileName);
			if (pctx->IsCancelled() || pctx->m_cbTotal > cbLimit)
			{
				// TCYE𒴂Ă̂Ŏ擾ȂBtH_
				if (pData->m_pShellFolder)
				{
					pData->m_pShellFolder = NULL;
				}
				if (pData->m_pStorage)
				{
					pData->m_pStorage = NULL;
				}
				pData->m_arrayProperties.Add(pctx->GetAbortText());
			}
			else
			{
				// f[^ă_~[f[^ƒu
				shared_ptr<CCBData> pData1;
				if (pData->m_pShellFolder)
				{
					pData1 = CCBData::CreateFromIShellFolder(pctx, pData->m_pShellFolder);
				}
				else if (pData->m_pStorage)
				{
					pData1 = CCBData::CreateFromIStorage(pctx, pData->m_pStorage);
				}

				if (pData1)
				{
				}
				else
				{
					pData1 = CCBData::CreateErrorData(L"f[^擾ł܂B");
					pData1->m_nIconType = CCBData::icontypeFolder;
				}
				pData1->m_strFileName = pData->m_strFileName;
				pData1->m_strDispFileName = pData->m_strDispFileName;
				at(i) = pData1;	// Ōf[^͔j
			}
		}
	}
}

void CCBDataArray::CreateSubTreeFromIStorage(CCBDataCreateContext* pctx, IStorage* pStorage)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	ASSERT(empty());	// ܂ȂƂɂȂ

	CComPtr<IEnumSTATSTG> pEnum;
	if (SUCCEEDED(pStorage->EnumElements(0, NULL, 0, &pEnum)))
	{
		STATSTG ss;
		ULONG c;
		while (SUCCEEDED(pEnum->Next(1, &ss, &c)) && c)
		{
			// ORs[ĉ
			CString strName = ss.pwcsName;
			CoTaskMemFree(ss.pwcsName);

			pctx->SetName(strName);
			if (pctx->IsCancelled())
				break;

			shared_ptr<CCBData> pDataChild;

			switch (ss.type)
			{
			case STGTY_STORAGE:
				// placeholderƂă_~[ĂB
				pDataChild = CCBData::CreateDummyFolder(L"IStorage");
				break;
			case STGTY_STREAM:
				{
					CComPtr<IStream> pStreamChild;
					if (SUCCEEDED(pStorage->OpenStream(strName, NULL, STGM_READ, NULL, &pStreamChild)))
					{
						pDataChild = CCBData::CreateFromIStream(pctx, pStreamChild);
					}
					if (!pDataChild)
					{
						pDataChild = CCBData::CreateErrorData(L"f[^擾ł܂B");
						pDataChild->m_strType = L"IStream";
					}
				}
				break;
			case STGTY_LOCKBYTES:
				pDataChild = CCBData::CreateErrorData(L"ΉĂȂ^CvłB");
				pDataChild->m_strType = L"STGTY_LOCKBYTES";
				pDataChild->m_nIconType = CCBData::icontypeUnknown;
				break;
			case STGTY_PROPERTY:
				pDataChild = CCBData::CreateErrorData(L"ΉĂȂ^CvłB");
				pDataChild->m_strType = L"STGTY_PROPERTY";
				pDataChild->m_nIconType = CCBData::icontypeUnknown;
				break;
			default:
				pDataChild = CCBData::CreateErrorData(L"sȃ^CvłB");
				pDataChild->m_nIconType = CCBData::icontypeUnknown;
				break;
			}
			if (pDataChild)
			{
				pDataChild->m_strFileName = strName;
				pctx->m_cbTotal += pDataChild->m_cbMem;
				push_back(pDataChild);
			}
		}

		// TCYE𒴂ĂȂ΃TuXg[W荞Ń_~[tH_u
		ULONGLONG cbLimit = (ULONGLONG)pApp->m_nSubDirSizeLimit * 1024 * 1024;
		for (int i = 0; i < (int)size(); i++)
		{
			CCBData* pDataChild = at(i).get();
			if (pDataChild->m_bPlaceHolder)
			{
				pctx->SetName(pDataChild->m_strFileName);
				if (pctx->IsCancelled() || pctx->m_cbTotal > cbLimit)
				{
					pDataChild->m_arrayProperties.Add(pctx->GetAbortText());
				}
				else
				{
					CString strFileName = pDataChild->m_strFileName;
					shared_ptr<CCBData> pDataChildNew;

					CComPtr<IStorage> pStorageChild;
					if (SUCCEEDED(pStorage->OpenStorage(strFileName, NULL, STGM_READ, NULL, 0, &pStorageChild)))
					{
						pDataChildNew = CCBData::CreateFromIStorage(pctx, pStorageChild);
					}
					if (!pDataChildNew)
					{
						pDataChildNew = CCBData::CreateErrorData(L"f[^擾ł܂B");
						pDataChildNew->m_strType = L"IStorage";
						pDataChildNew->m_nIconType = CCBData::icontypeFolder;
					}
					pDataChildNew->m_strFileName = strFileName;
					at(i) = pDataChildNew;
				}
			}
		}
	}
}

void CCBDataArray::CreateSubTreeFromIShellFolder(CCBDataCreateContext* pctx, IShellFolder* pFolder)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	ASSERT(empty());

	CComPtr<IEnumIDList> pEnum;
	if (SUCCEEDED(pFolder->EnumObjects(NULL, 
		SHCONTF_FOLDERS | SHCONTF_NONFOLDERS | SHCONTF_INCLUDEHIDDEN | SHCONTF_INCLUDESUPERHIDDEN,
		&pEnum)))
	{
		while (true) 
		{
			ULONG c;
			CComHeapPtr<ITEMIDLIST> pidl;

			if (!SUCCEEDED(pEnum->Next(1, &pidl, &c))) {
				break;
			}
			if (!c) {
				break;
			}

			STRRET strretFile;
			CString strFileName;
			CString strDispFileName;
			if (SUCCEEDED(pFolder->GetDisplayNameOf(pidl, SHGDN_FORPARSING | SHGDN_INFOLDER, &strretFile)))
			{
				strFileName = GetResultAndFree(&strretFile);
			}
			if (SUCCEEDED(pFolder->GetDisplayNameOf(pidl, SHGDN_NORMAL, &strretFile)))
			{
				strDispFileName = GetResultAndFree(&strretFile);
			}

			pctx->SetName(strFileName);
			if (pctx->IsCancelled())
			{
				break;
			}

			shared_ptr<CCBData> pDataChild;
			CComPtr<IShellFolder> pSubFolder;
			CComPtr<IStream> pStream;

			if (SUCCEEDED(pFolder->BindToObject(pidl, NULL, IID_IStream, (void**)&pStream)))
			{
				pDataChild = CCBData::CreateFromIStream(pctx, pStream);
				pctx->m_cbTotal += pDataChild->m_cbMem;
			}
			else if (SUCCEEDED(pFolder->BindToObject(pidl, NULL, IID_IShellFolder, (void**)&pSubFolder)))
			{
				// _~[f[^ĂBTCYOKȂ牺Ŏ擾
				pDataChild = CCBData::CreateDummyFolder(L"IShellFolder");
				pDataChild->m_pShellFolder = pSubFolder;
				// pShellFolder͉̉
			}
			else
			{
				pDataChild = make_shared<CCBData>();
				pDataChild->m_strType = L"s";
				pDataChild->m_nIconType = CCBData::icontypeUnknown;
			}

			if (pDataChild)
			{
				pDataChild->m_strFileName = strFileName;
				pDataChild->m_strDispFileName = strDispFileName;
				push_back(pDataChild);
			}
		}

		// TCYE𒴂ĂȂ΃TutH_
		ULONGLONG cbLimit = (ULONGLONG)pApp->m_nSubDirSizeLimit * 1024 * 1024;
		for (int i = 0; i < (int)size(); i++)
		{
			CCBData* pDataChild = at(i).get();
			if (pDataChild->m_bPlaceHolder)
			{
				pctx->SetName(pDataChild->m_strFileName);
				if (pctx->IsCancelled() || pctx->m_cbTotal > cbLimit)
				{
					// TCY𒴂Ă̂Ŏ擾ȂBtH_
					if (pDataChild->m_pShellFolder)
					{
						pDataChild->m_pShellFolder = NULL;
					}
					pDataChild->m_arrayProperties.Add(pctx->GetAbortText());
				}
				else
				{
					// f[^ă_~[f[^ƒu
					shared_ptr<CCBData> pDataChild1 = CCBData::CreateFromIShellFolder(pctx, pDataChild->m_pShellFolder);
					if (pDataChild1)
					{
					}
					else
					{
						pDataChild1 = CCBData::CreateErrorData(L"f[^擾ł܂B");
						pDataChild1->m_strType = L"IShellFolder";
						pDataChild1->m_nIconType = CCBData::icontypeFolder;
					}
					pDataChild1->m_strFileName = pDataChild->m_strFileName;
					pDataChild1->m_strDispFileName = pDataChild->m_strDispFileName;
					at(i) = pDataChild1;
				}
			}
		}
	}
}

void CCBDataArray::CreateSubTreeFromDirectory(CCBDataCreateContext* pctx, const CString& strDir)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	ASSERT(empty());

	CString strPathNameToSearch = MakePathName(strDir, L"*");

	WIN32_FIND_DATA wfd;
	HANDLE hFind = FindFirstFile(strPathNameToSearch, &wfd);
	if (hFind != INVALID_HANDLE_VALUE)
	{
		do
		{
			shared_ptr<CCBData> pDataChild;
			CString strFileName = wfd.cFileName;
			pctx->SetName(strFileName);
			if (pctx->IsCancelled())
				break;


			if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				if (strFileName == L"." || strFileName == L"..")
				{
				}
				else
				{
					// v[Xz_Ă
					pDataChild = CCBData::CreateDummyFolder(L"fBNg");
				}
			}
			else
			{
				pDataChild = CCBData::CreateFromFile(pctx, MakePathName(strDir, strFileName));
				if (pDataChild)
				{
				}
				else
				{
					pDataChild = CCBData::CreateErrorData(L"f[^荞߂܂");
					pDataChild->m_strType = L"fBXNt@C";
				}
			}
			if (pDataChild)
			{
				pDataChild->m_strFileName = strFileName;
				pctx->m_cbTotal += pDataChild->m_cbMem;
				push_back(pDataChild);
			}
		} while (FindNextFile(hFind, &wfd));

		FindClose(hFind);
	}
	// TCYE𒴂ĂȂ΃TufBNg
	ULONGLONG cbLimit = (ULONGLONG)pApp->m_nSubDirSizeLimit * 1024 * 1024;
	for (int i = 0; i < (int)size(); i++)
	{
		CCBData* pDataChild = at(i).get();
		if (pDataChild->m_bPlaceHolder)
		{
			pctx->SetName(pDataChild->m_strFileName);
			if (pctx->IsCancelled() || pctx->m_cbTotal > cbLimit)
			{
				pDataChild->m_arrayProperties.Add(pctx->GetAbortText());
			}
			else
			{
				CString strSubDir = pDataChild->m_strFileName;
				shared_ptr<CCBData> pDataChildNew;
				pDataChildNew = CCBData::CreateFromDirectory(pctx, MakePathName(strDir, strSubDir));
				if (!pDataChildNew)
				{
					pDataChildNew = CCBData::CreateErrorData(L"tf[^擾ł܂B");
					pDataChildNew->m_strType = L"fBNg";
					pDataChildNew->m_nIconType = CCBData::icontypeFolder;
				}
				pDataChildNew->m_strFileName = strSubDir;
				at(i) = pDataChildNew;
			}
		}
	}
}

void CCBDataArray::CreateSubTreeFromHDROP(CCBDataCreateContext* pctx, HDROP hDrop)
{
	CClipMMApp* pApp = (CClipMMApp*)AfxGetApp();

	ASSERT(empty());

	BOOL bSuccess = TRUE;
	UINT nCount = DragQueryFileCount(hDrop);
	for (UINT i = 0; i < nCount; i++)
	{
		shared_ptr<CCBData> pData;
		CString strFileName;
		if (DragQueryFile(hDrop, i, strFileName))
		{
			pctx->SetName(strFileName);
			if (pctx->IsCancelled())
				break;


			DWORD dwFileAttributes = GetFileAttributes(strFileName);
			if (dwFileAttributes != INVALID_FILE_ATTRIBUTES) 
			{
				if (dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) 
				{
					// placeholderƂă_~[ĂB
					pData = CCBData::CreateDummyFolder(L"fBNg");
				}
				else
				{
					pData = CCBData::CreateFromFile(pctx, strFileName);
					if (!pData)
					{
						pData = CCBData::CreateErrorData(L"f[^擾ł܂B");
						pData->m_strType = L"fBXNt@C";
					}
				}
			}
			else
			{
				pData = CCBData::CreateErrorData(L"t@C܂B");
			}

			if (pData)
			{
				pData->m_strFileName = strFileName;
			}
		}
		else
		{
			pData = CCBData::CreateErrorData(L"t@C擾ł܂B");
		}
		if (pData)
		{
			pData->m_strFileName = strFileName;
			pctx->m_cbTotal += pData->m_cbMem;
			push_back(pData);
		}
	}
			
	// TCYE𒴂ĂȂ΃TufBNg
	ULONGLONG cbLimit = (ULONGLONG)pApp->m_nSubDirSizeLimit * 1024 * 1024;
	for (int i = 0; i < (int)size(); i++)
	{
		CCBData* pData = at(i).get();
		if (pData->m_bPlaceHolder)
		{
			pctx->SetName(pData->m_strFileName);
			if (pctx->IsCancelled() || pctx->m_cbTotal > cbLimit)
			{
				pData->m_arrayProperties.Add(pctx->GetAbortText());
			}
			else
			{
				CString strDirName = pData->m_strFileName;

				shared_ptr<CCBData> pDataNew;
				pDataNew = CCBData::CreateFromDirectory(pctx, strDirName);
				if (!pDataNew)
				{
					pDataNew = CCBData::CreateErrorData(L"f[^擾ł܂B");
					pDataNew->m_strType = L"fBNg";
					pDataNew->m_nIconType = CCBData::icontypeFolder;
				}
				pDataNew->m_strFileName = strDirName;
				at(i) = pDataNew;
			}
		}
	}
}

void CCBDataArray::CreateSubTreeFromFileDescriptor(CCBDataCreateContext* pctx, CCBData_FileDescriptor* pDataDescriptor, FORMATETC& fmtetcContents)
{
	ASSERT(empty());

	// pData̎qǂɓ
	for (int n = 0; n < pDataDescriptor->m_arrayFileNames.GetSize(); n++)
	{
		pctx->SetName(pDataDescriptor->m_arrayFileNames[n]);
		if (pctx->IsCancelled())
			break;

		FORMATETC fmtetc = fmtetcContents;
		fmtetc.lindex = n;
		pctx->m_cbTotal = 0;
		shared_ptr<CCBData> pDataChild = CCBData::CreateSub(pctx, fmtetc, TYMED_HGLOBAL);
		if (pDataChild)
		{
			pDataChild->m_strFileName = pDataDescriptor->m_arrayFileNames[n];
			push_back(pDataChild);
		}
	}
}
